/*
 * Tinity Plugin
 *
 * Copyright (C) 2009 Christian Franke <nobody@nowhere.ws>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef PURPLE_PLUGINS
#define PURPLE_PLUGINS
#endif
#define PLUGIN_ID "core-cfchris6-tinity"
#define PLUGIN_TITLE "Tinity"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#include <curl/curl.h>
#include <glib.h>

#include "conversation.h"
#include "debug.h"
#include "plugin.h"
#include "version.h"
#include "pluginpref.h"
#include "prefs.h"
#include "signals.h"

#define post_url \
    purple_prefs_get_string("/plugins/core/cfchris6_tinity/url")

struct recvStruct {
    char *buffer;
    size_t datasize;
    size_t bufsize;
};

struct writeStruct {
    PurpleConversation *conversation;
    PurpleAccount *account;
    char *sender;
    char *buffer;
};

static PurplePlugin *tinity_plugin = NULL;
static GRegex *url_regex = NULL;
static GRegex *xmlns_cleaner = NULL;

static void recvStruct_init(struct recvStruct *dat) {
    dat->buffer = NULL;
    dat->bufsize = 0;
    dat->datasize = 0;
}

static size_t write_to_char(void *buffer, size_t size, size_t nmemb,
        void *userp) {
    struct recvStruct *myData = userp;
    if (myData->bufsize*2 < myData->datasize+size*nmemb) {
        if (myData->datasize+size*nmemb <= 2036) {
            myData->bufsize = 2036;
        } else if (myData->datasize+size*nmemb <= 4084) {
            myData->bufsize = 4084;
        } else if ((myData->datasize+size*nmemb - 4084)%2048 == 0) {
            myData->bufsize = myData->datasize+size*nmemb;
        } else {
            myData->bufsize = 4084 + 2048
                + 2048 * ((myData->datasize+size*nmemb - 4084)/2048);
        }
        myData->buffer = g_realloc(myData->buffer, myData->bufsize*2);
    }
    g_memmove(myData->buffer + myData->datasize, buffer, size*nmemb);
    myData->datasize += size*nmemb;
    return size*nmemb;
}

static gchar * create_short_url(gchar *url) {
    CURL *chandle;
    gchar *req_url;
    struct recvStruct myRecv;

    recvStruct_init(&myRecv);
    chandle = curl_easy_init();

    req_url = g_malloc(strlen(post_url) + strlen("?url=")
            + strlen(url) + 1);
    g_sprintf(req_url, "%s?url=%s", post_url, url);
    curl_easy_setopt(chandle, CURLOPT_URL, req_url);
    g_free(req_url);
    curl_easy_setopt(chandle, CURLOPT_WRITEFUNCTION, write_to_char);
    curl_easy_setopt(chandle, CURLOPT_WRITEDATA, &myRecv);

    curl_easy_perform(chandle);
    curl_easy_cleanup(chandle);

    myRecv.buffer = g_realloc(myRecv.buffer, myRecv.datasize + 1);
    myRecv.buffer[myRecv.datasize] = '\0';
    return myRecv.buffer;
}

static gboolean write_to_im(gpointer data) {
    struct writeStruct *myWrite = data;

    if(!myWrite->conversation) {
        myWrite->conversation =
            purple_find_conversation_with_account(PURPLE_CONV_TYPE_IM,
                    myWrite->sender, myWrite->account);
    }

    if(!myWrite->conversation) {
        g_free(myWrite->sender);
        g_free(myWrite);
        return FALSE;
    }
    purple_conv_im_write(myWrite->conversation->u.im, NULL,
            myWrite->buffer, PURPLE_MESSAGE_SYSTEM, time(NULL));
    g_free(myWrite->buffer);
    g_free(myWrite->sender);
    g_free(myWrite);
    return FALSE;
}

static void received_im_msg_cb(PurpleAccount *account, char *sender,
        char *msg, PurpleConversation *conv, PurpleMessageFlags flags,
        void *data) {
    GMatchInfo *info;
    gchar *msg_stripped;

    if (flags & (PURPLE_MESSAGE_SYSTEM | PURPLE_MESSAGE_NO_LINKIFY
                  | PURPLE_MESSAGE_INVISIBLE) > 0)
        return;
    g_regex_ref(xmlns_cleaner);
    msg_stripped = g_regex_replace(xmlns_cleaner, msg, -1, 0, "", 0, NULL);
    g_regex_unref(xmlns_cleaner);
    g_regex_ref(url_regex);
    g_regex_match(url_regex, msg_stripped, 0, &info);
    while(g_match_info_matches(info)) {
        gchar *url = g_match_info_fetch(info,1);
        if (strlen(url) > 22) {
            gchar *short_url = create_short_url(url);
            struct writeStruct *writeInfo =
                g_malloc(sizeof(struct writeStruct));
            writeInfo->conversation = conv;
            writeInfo->account = account;
            writeInfo->sender = g_malloc(strlen(sender) + 1);
            g_memmove(writeInfo->sender, sender, strlen(sender) + 1);
            writeInfo->buffer = g_malloc(strlen("Short url for ")
                    + strlen(url) + strlen(" is ") + strlen(short_url) + 1);
            g_sprintf(writeInfo->buffer, "Short url for %s is %s", url,
                    short_url);
            g_free(short_url);
            purple_timeout_add(20, write_to_im, writeInfo);
        }
        g_free(url);
        g_match_info_next(info, NULL);
    }
    g_regex_unref(url_regex);
    g_free(msg_stripped);
}

static gboolean
plugin_load(PurplePlugin *plugin) {
    char *myUrl = NULL;
    tinity_plugin = plugin;
    void *conv_handle = purple_conversations_get_handle();

    curl_global_init(CURL_GLOBAL_ALL);

    url_regex = g_regex_new("(https?://.*?)(?:[ '\"\t\n]|$)",
            G_REGEX_OPTIMIZE, 0, NULL);
//    xmlns_cleaner = g_regex_new("xmlns=[\"']?.*?(?:[\"' \t\n]|$)",
//            G_REGEX_OPTIMIZE, 0, NULL);
    // Remove all tags till redesign of received_im_msg_cb
    xmlns_cleaner = g_regex_new("<.*?>", G_REGEX_OPTIMIZE, 0, NULL);

    purple_signal_connect(conv_handle, "received-im-msg",
            plugin, PURPLE_CALLBACK(received_im_msg_cb), NULL);

    return TRUE;
}

static gboolean
plugin_unload(PurplePlugin *plugin) {
    if (url_regex) {
        g_regex_unref(url_regex);
        url_regex = NULL;
    }
    if (xmlns_cleaner) {
        g_regex_unref(xmlns_cleaner);
        xmlns_cleaner = NULL;
    }

    tinity_plugin = NULL;
    return TRUE;
}

static PurplePluginPrefFrame *
get_plugin_pref_frame(PurplePlugin *plugin) {
    PurplePluginPrefFrame *frame;
    PurplePluginPref *ppref;

    frame = purple_plugin_pref_frame_new();

    ppref = purple_plugin_pref_new_with_name_and_label(
            "/plugins/core/cfchris6_tinity/url", "Create-Url:");
    purple_plugin_pref_frame_add(frame, ppref);
    return frame;
}

static PurplePluginUiInfo prefs_info = {
    get_plugin_pref_frame,
    0,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL
};

static PurplePluginInfo info = {
    PURPLE_PLUGIN_MAGIC,
    PURPLE_MAJOR_VERSION,
    PURPLE_MINOR_VERSION,
    PURPLE_PLUGIN_STANDARD,
    NULL,
    0,
    NULL,
    PURPLE_PRIORITY_DEFAULT,
    PLUGIN_ID,
    PLUGIN_TITLE,
    PLUGIN_VERSION,
    "Finds incoming urls and makes them short",
    "This plugin uses a service like http://tinyurl.com to make "
    "incoming urls shorter.\n\n"
    "Warning:\n"
    "Of course, all urls you shortened will be known to the service "
    "you are using, so it might be a good idea to set up your own "
    "\"Short-Url-Service\". (the url used for shortening can be set "
    "in the plugin configuration)",
    "Christian Franke <nobody@nowhere.ws>",
    "http://nowhere.ws/tinity/",
    plugin_load,
    plugin_unload,
    NULL,
    NULL,
    NULL,
    &prefs_info,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL
};

static void
init_plugin(PurplePlugin *plugin)
{
    purple_prefs_add_none("/plugins/core/cfchris6_tinity");
    purple_prefs_add_string("/plugins/core/cfchris6_tinity/url",
                            "http://tinyurl.com/api-create.php");
}

PURPLE_INIT_PLUGIN(core-cfchris6-tinity, init_plugin, info)
